
import numpy as np
from scipy.integrate import quad


def system():
    k = 1.8e6        # stiffness
    omega_n = 2.1  # eigenfrequency
    zeta = 0.01   # damping ratio
    ###
    return omega_n, zeta, k


def load_par():
    AR = [0]
    sigma_w = 1e5
    F0 = []
    omega_f = []
    return np.array(AR), sigma_w, np.array(F0), np.array(omega_f)


def discretization():
    dt = .1
    n_F = 10
    n_r = 201
    n_theta = 0
    psi = 1
    xi = 1
    lambda_f = 3.
    lambda_r = 3
    return dt, n_F, n_r, n_theta, psi, xi, lambda_f, lambda_r




def mesh():
    dt, n_F, n_r, n_theta, psi, xi, lambda_f, lambda_r = discretization()
    AR, sigma_w, F0, omega_f = load_par()
    omega_n, zeta, k = system()
    mesh = {}
    if n_F > 0:

        if len(AR) == 2:
            phi1 = AR[0]
            phi2 = AR[1]
            var_F = (1. - phi2) / (1. + phi2) * sigma_w**2 / ((1. - phi2)**2 - phi1**2)
            sigma_f = var_F**.5
            F = np.linspace(-lambda_f * sigma_f, lambda_f * sigma_f, n_F)
            mesh['F2'] = F
        else:
            sigma_f = sigma_w / (1. - AR[0]**2)**.5
            F = np.linspace(-lambda_f * sigma_f, lambda_f * sigma_f, n_F)
        mesh['F'] = F
    else:
        F = 0

    if n_theta > 0:
        phi = np.linspace(0, 2 * np.pi, n_theta, endpoint=False)
        dphi = phi[1] - phi[0]
        phi += dphi / 2
        mesh['phi'] = phi
        if len(F0) == 2:
            n_theta2 = n_theta
            phi = np.linspace(0, 2 * np.pi, n_theta2, endpoint=False)
            dphi = phi[1] - phi[0]
            phi += dphi / 2
            mesh['phi2'] = phi
    else:
        phi = 0
    x = np.linspace(-.5, .5, n_r)  # displacement and velocity must be defined here
    v = np.linspace(-1, 1, n_r)
    mesh['x'] = x
    mesh['v'] = v
    return mesh


def par():
    AR, sigma_w, F0, omega_f = load_par()
    omega_n, zeta, k = system()
    dt, n_F, n_r, n_theta, psi, xi, lambda_f, lambda_r = discretization()
    par = {'k': k, 'omega_n': omega_n, 'zeta': zeta, 'dt': dt}
    if len(AR) > 0:
        par['AR'] = AR
        par['sigma_w'] = sigma_w
    if len(F0) > 0:
        par['F0'] = F0
        par['omega_f'] = omega_f
    return par


def load_case(par):
    if 'AR' in par:
        nAR = len(par['AR'])
    else:
        nAR = 0
    if 'F0' in par:
        nHAR = len(par['F0'])
    else:
        nHAR = 0
    dim = nHAR + nAR
    if dim == 1:
        if nAR == 1:
            return 1
        else:
            return 2
    elif dim == 2:
        if nAR == 2:
            return 3
        elif nAR == 1:
            return 4
        else:
            return 5
    elif dim == 3:
        if nAR == 3:
            return 6
        elif nAR == 2:
            return 7
        elif nAR == 1:
            return 8
        else:
            return 9


def AR_NOISE():
    dt, n_F, n_r, n_theta, psi, xi, lambda_F, lambda_r = discretization()
    AR, sigma_w, F0, omega_f = load_par()
    if len(AR) == 1:
        sigma_F = sigma_w / (1. - AR[0]**2)**.5
    elif len(AR) == 2:
        phi1 = AR[0]
        phi2 = AR[1]
        var_F = (1. - phi2) / (1. + phi2) * sigma_w**2 / ((1. - phi2)**2 - phi1**2)
        sigma_F = var_F**.5

    a = (8. * psi**2 * lambda_F**2 / n_F**2 * (sigma_F**2 / sigma_w**2) - 2.) * xi
    b = 2. * psi * lambda_F * sigma_F / n_F
    W = b * np.array([-1, 0, 1])
    pW = 1. / (2 + a) * np.array([1, a, 1])
    return W, pW


def response_displ(omega, dt, AR1, sigma_w):
    omega_n, zeta, k = system()
    beta = omega / omega_n
    return sigma_w**2 * dt / np.pi / (1. - 2. * AR1 * np.cos(omega * dt) + AR1**2) / (k**2 * ((1. - beta**2)**2 + (2 * zeta * beta)**2))


def response_vel(omega, dt, AR1, sigma_w):
    omega_n, zeta, k = system()
    beta = omega / omega_n
    return omega**2 * sigma_w**2 * dt / np.pi / (1. - 2. * AR1 * np.cos(omega * dt) + AR1**2) / (k**2 * ((1. - beta**2)**2 + (2 * zeta * beta)**2))

def FRF(omega_f):
    omega_n, zeta, k = system()
    beta = omega_f / omega_n
    return 1. / (k * ((1. - beta**2)**2 + (2 * zeta * beta)**2)**.5)

