# -*- coding: utf-8 -*-
"""
Created on Fri Feb 10 12:24:58 2017

@author: Odd Eiken
"""

import numpy as np
from numba import guvectorize, jit
import time
from scipy.sparse import csc_matrix
import cell_discretization as cell
import scipy.stats as ss


@guvectorize(
    'void(f8[:], f8[:], f8[:], f8[:], f8[:], f8[:], f8[:], f8[:])',
    '(m),(n),(n),(n),(n),(n)->(n),(n)',
    target='parallel')
def RKN(par, x, v, F0, F1, F2, xn, vn):
    omega_n, zeta, k, h = par
    K = 0.5 * h * omega_n**2
    for i in xrange(len(x)):
        k1 = K * (F0[i] / k - 2 * zeta / omega_n * v[i] - x[i])
        k2 = K * (
            F1[i] / k -
            2 * zeta / omega_n * (v[i] + k1) -
            (x[i] + 0.5 * h * (v[i] + .5 * k1)))
        k3 = K * (
            F1[i] / k -
            2 * zeta / omega_n * (v[i] + k2) -
            (x[i] + 0.5 * h * (v[i] + .5 * k1)))
        k4 = K * (
            F2[i] / k -
            2 * zeta / omega_n * (v[i] + 2 * k3) -
            (x[i] + h * (v[i] + k3)))
        xn[i] = x[i] + h * (v[i] + (k1 + k2 + k3) / 3.)
        vn[i] = v[i] + (k1 + 2 * k2 + 2 * k3 + k4) / 3.


@jit('f8[:](i8[:],i8)', nopython=True)
def create_random(X, no_points):
    '''
    Creating random monte carlo points
    input:
    -----
    X : array(integers)
    no_points : integer number of monte carlo points for each cell
    X_rand : random points array(float), size is no_points time larger than X,
    and must be defined beforehand
    '''
    n = X.size
    X_rand = np.zeros(n * no_points, dtype=np.float64)
    r = np.empty(no_points, dtype=np.float64)
    for i in xrange(len(X)):
        for j in xrange(no_points):
            r[j] = np.random.rand() - 0.5
        X_rand[no_points * i:no_points * i + no_points] = X[i] + r
    return X_rand


@guvectorize('void(i8[:], f8[:], f8[:], f8[:])', '(n), (m), (n) -> (n)', target='parallel')
def index_to_value(X_idx, x, rand, X):
    dx = x[1] - x[0]
    for i in xrange(len(X_idx)):
        X[i] = x[X_idx[i]] + dx * rand[i]


@guvectorize('void(f8[:], f8, i8[:])', '(n),()->(n)', target='cpu')
def find_index(values, dx, res):
    for i in xrange(len(values)):
        res[i] = round((values[i] - dx / 2) / dx)


@guvectorize('void(i8, i8, i8[:], i8[:], i8[:], i8[:])', '(), (), (n) ->(n), (n), (n)', target='parallel')
def argwhere3d(len_v, len_f, states, X_idx, V_idx, F_idx):
    '''
    from cell number to state indexes
    '''
    for i in xrange(len(states)):
        X_idx[i] = states[i] / int(len_v * len_f)
        V_idx[i] = (states[i] - X_idx[i] * len_v * len_f) / int(len_f)
        F_idx[i] = (states[i] - X_idx[i] * len_v * len_f - V_idx[i] * len_f)


@guvectorize('void(i8, i8, i8, i8[:], i8[:], i8[:], i8[:], i8[:])', '(), (), (), (n) ->(n), (n), (n), (n)', target='parallel')
def argwhere4d(len_v, len_f1, len_f2, states, X_idx, V_idx, F1_idx, F2_idx):
    '''
    from cell number to state indexes
    '''
    for i in xrange(len(states)):
        X_idx[i] = states[i] / int(len_v * len_f1 * len_f2)
        V_idx[i] = (states[i] - X_idx[i] * len_v * len_f1 * len_f2) / int(len_f1 * len_f2)
        F1_idx[i] = (states[i] - X_idx[i] * len_v * len_f1 * len_f2 - V_idx[i] * len_f1 * len_f2) / int(len_f2)
        F2_idx[i] = (states[i] - X_idx[i] * len_v * len_f1 * len_f2 - V_idx[i] * len_f1 * len_f2 - F1_idx[i] * len_f2)
      
      
      
@guvectorize('void(i8, i8, i8, i8, i8[:], i8[:], i8[:], i8[:], i8[:], i8[:])', '(), (), (), (), (n) ->(n), (n), (n), (n), (n)', target='parallel')
def argwhere5d(len_v, len_f1, len_f2, len_f3, states, X_idx, V_idx, F1_idx, F2_idx, F3_idx):
    '''
    from cell number to state indexes
    '''
    for i in xrange(len(states)):
        X_idx[i] = states[i] / int(len_v * len_f1 * len_f2 * len_f3)
        V_idx[i] = (states[i] - X_idx[i] * len_v * len_f1 * len_f2 * len_f3) / int(len_f1 * len_f2 * len_f3)
        F1_idx[i] = (states[i] - X_idx[i] * len_v * len_f1 * len_f2 * len_f3 - V_idx[i] * len_f1 * len_f2 * len_f3) / int(len_f2 * len_f3)
        F2_idx[i] = (states[i] - X_idx[i] * len_v * len_f1 * len_f2 * len_f3 - V_idx[i] * len_f1 * len_f2 * len_f3 - F1_idx[i] * len_f2 * len_f3) / int(len_f3)
        F3_idx[i] = (states[i] - X_idx[i] * len_v * len_f1 * len_f2 * len_f3 - V_idx[i] * len_f1 * len_f2 * len_f3 - F1_idx[i] * len_f2 * len_f3 - F2_idx[i] * len_f3) 





@jit('i4(f8)', nopython=True)
def H(x):
    '''
    Heavside step function: returns 0 if x<0, and 1 if x>=0
    '''
    if x >= 0:
        return 1
    else:
        return 0


@guvectorize(
    'void(f8[:], f8[:], f8, f8, f8[:], f8[:])',
    '(n),(n),(),()->(n),(n)')
def Wall(x, v, xlim, vlim, xn, vn):
    for i in xrange(len(xn)):
        if xn[i]**2 > xlim**2 or vn[i]**2 > vlim**2:
            L = max(
                H(abs(xn[i]) - xlim) * (abs(xn[i]) - xlim) / abs(xn[i] - x[i]),
                H(abs(vn[i]) - vlim) * (abs(vn[i]) - vlim) / abs(vn[i] - v[i]))
            xn[i] = x[i] + (xn[i] - x[i]) * (1 - L)
            vn[i] = v[i] + (vn[i] - v[i]) * (1 - L)


@guvectorize(
    'void(f8[:], f8[:], f8[:], f8[:])',
    '(n),(n)->(n),(n)')
def Wall_zero_vel(x, v, xn, vn):
    for i in xrange(len(x)):
        if vn[i] / v[i] < 0:
            xn[i] = (x[i] - xn[i]) / (v[i] - vn[i]) * v[i] + x[i]
            vn[i] = 0


@jit('f8[:](f8, f8[:])')
def limit_F(F_lim, f):
    for i in xrange(f.size):
        if abs(f[i]) > F_lim:
            f[i] = f[i] / abs(f[i]) * F_lim
    return f


def ARForce_mapping(mode=1):
    '''
    mode:
    -----
    1 : plots and returns probability distribution of F: returns forcevector and probability vector (F, p)
    2 : finds a value for psi: returns psi values and square mean deviation (psi, e)
    '''
    import matplotlib.pyplot as plt
    from scipy.interpolate import interp1d
    iterations = 50
    AR, sigma_w, F0, phi = cell.load_par()
    AR1 = AR[0]
    try:
        AR2 = AR[1]
        sigma_F = ((1. - AR2) / (1 - + AR2) * sigma_w**2 / ((1 - AR2)**2 - AR1**2))**.5
    except:
        IndexError
        AR2 = 0
        sigma_F = sigma_w / (np.sqrt(1. - AR1**2))
    dt, n_F, n_r, n_phi, psi, xi, lambda_F, lambda_r = cell.discretization()
    F = np.linspace(- lambda_F * sigma_F, lambda_F * sigma_F, n_F)
    dF = F[1] - F[0]
    if mode == 2:
    	RP = 10000
    	Rand = (np.random.rand(RP) - .5) * dF
        psi_min = max([n_F * (1 - AR1**2)**.5 / 2. / lambda_F, 1])
        psi_max = 1.8
        len_psi = iterations
        psi = np.linspace(psi_min, psi_max, len_psi)
        y = np.empty(len_psi)
        # xi = 1.
        fig, (ax1, ax2) = plt.subplots(2)
    elif mode == 1:
    	RP = 100000
    	Rand = (np.random.rand(RP) - .5) * dF
        psi = [psi]
        len_psi = 1
        fig, ax = plt.subplots(figsize=(6 * 1.5, 4 * 1.5))
    for i in range(len_psi):
        print i
        P = np.zeros((n_F, n_F))
        a = (8. * psi[i]**2 * lambda_F**2 / (n_F**2 * (1. - AR[0]**2)) - 2.) * xi
        b = 2. * psi[i] * lambda_F * sigma_w / (n_F * np.sqrt(1. - AR[0]**2)) 
        W = b * np.array([-1, 0, 1])
        pW = 1. / (2 + a) * np.array([1, a, 1])
        for F_idx in xrange(n_F):
            for W_idx in xrange(3):
                F0 = np.repeat(F[F_idx], RP) + Rand
                F1 = F0 * AR1 + W[W_idx]
                flim = F[-1] + dF / 2 * .99
                F1 = limit_F(flim, F1)
                F1_idx = find_index(F1 + dF / 2 * (n_F % 2), dF) + n_F / 2
                for m in xrange(RP):
                    P[F1_idx[m], F_idx] += 1. * pW[W_idx] / float(RP)
        M = np.matrix(P)
        p = np.array((M**100000).T[0])[0]
        f = np.linspace(- lambda_F * sigma_F, lambda_F * sigma_F, 1000)
        pF = ss.norm.pdf(f, loc=0, scale=sigma_F)
        if mode == 1:
        	print 'alpha = ', 8. * psi[i]**2 * lambda_F**2 / (n_F**2 * (1. - AR[0]**2)) - 2.
    		ax.plot(F, p / dF, '.-', c='b',  label='$\psi$ = %s' % (psi[i]))
    		ax.plot(F, p / dF, '.-', c='b', label=r'$\xi$ = %s' % (xi))
        	ax.plot(f, pF, label='exact', c='k')
        	plt.legend(fancybox=True)
        	plt.show()
        	return F, p
        Fc = interp1d(F, p / dF, kind='linear')
        error = np.sum((Fc(f) - pF)**2)**.5
        y[i] = error
    Rd = y
    R = interp1d(psi, Rd, kind='cubic')
    x = np.linspace(psi[0], psi[-1], 1000)
    psi_opt = np.argmin(R(x))
    psi_opt = round(x[psi_opt], 2)
    

    print 'optimal choice of $\psi$ =', psi_opt
    print 'alpha = ', (8. * psi_opt**2 * lambda_F**2 / (n_F**2 * (1. - AR[0]**2)) - 2.)
    ax1.plot(x, R(x), '-')
    ax1.plot(psi, Rd, '.')
    ax1.plot(psi_opt, R(psi_opt), 'ro', label=r'$\psi = $%s' %(psi_opt))
    ax1.set_xlabel('$\psi$', size=15)
    ax1.set_ylabel('Residual', size=14)

    xi = np.linspace(1.4,2.5,iterations)
    y = np.zeros(len(xi))
    psi = psi_opt
    for i in range(len(xi)):
        a = (8. * psi**2 * lambda_F**2 / (n_F**2 * (1. - AR[0]**2)) - 2.) * xi[i]
        b = 2. * psi * lambda_F * sigma_w / (n_F * np.sqrt(1. - AR[0]**2)) 
        W = b * np.array([-1, 0, 1])
        pW = 1. / (2 + a) * np.array([1, a, 1])
        print i
        P = np.zeros((n_F, n_F))
        for F_idx in xrange(n_F):
            for W_idx in xrange(3):
                F0 = np.repeat(F[F_idx], RP) + Rand
                F1 = F0 * AR1 + W[W_idx]
                flim = F[-1] + dF / 2 * .99
                F1 = limit_F(flim, F1)
                F1_idx = find_index(F1 + dF / 2 * (n_F % 2), dF) + n_F / 2
                for m in xrange(RP):
                    P[F1_idx[m], F_idx] += 1. * pW[W_idx] / float(RP)
        M = np.matrix(P)
        p = np.array((M**100000).T[0])[0]
        Fc = interp1d(F, p / dF, kind='linear')
        P = Fc(f)
        P = P/np.sum(P)
        std_F_hat = np.sum(f**2*P)**.5
        y[i] = std_F_hat
    Rd = (y - sigma_F)**2
    R = interp1d(xi, Rd, kind='cubic')
    x = np.linspace(xi[0], xi[-1], 1000)
    xi_opt = np.argmin(R(x))
    xi_opt = round(x[xi_opt], 2)
    print 'optimal choice of $xi$ =', xi_opt
    ax2.plot(x, R(x), '-')
    ax2.plot(xi, Rd, '.')
    ax2.plot(xi_opt, R(xi_opt), 'ro', label=r'$\xi = $%s' %(xi_opt))
    ax2.set_xlabel(r'$\xi$', size=15)
    ax2.set_ylabel('Residual', size=14)
    ax2.legend(fancybox=True)
    ax1.legend(fancybox=True)
    plt.tight_layout()
    plt.show()


def AR2Force_mapping(mode, K=2):
    '''
    mode:
    -----
    1 : plots and returns probability distribution of F: returns forcevector and probability vector (F, p)
    2 : finds a value for psi: returns psi values and square mean deviation (psi, e)
    K : (psi, e) chosen between 1/K and K times current value
    '''
    import matplotlib.pyplot as plt
    from scipy.interpolate import interp1d
    iterations = 8
    AR, sigma_w, F0, phi = cell.load_par()
    dt, n_F, n_r, n_phi, psi, xi, lambda_F, lambda_r = cell.discretization()
    AR1 = AR[0]
    AR2 = AR[1]
    sigma_F = ((1. - AR2) / (1. + AR2) * sigma_w**2 / ((1. - AR2)**2 - AR1**2))**.5
    F = np.linspace(-lambda_F * sigma_F, lambda_F * sigma_F, n_F)
    dF = F[1] - F[0]

    if mode == 2:
        RP = 10000
        Rand = (np.random.rand(RP) - .5) * dF
        psi_min = max([n_F / 2. / lambda_F * sigma_w / sigma_F, psi / K])
        psi_max = psi * K
        len_psi = iterations
        psi = np.linspace(psi_min, psi_max, len_psi)
        y = np.empty(len_psi)
        # xi = 1.
        fig, (ax1, ax2) = plt.subplots(2)
    elif mode == 1:
        if psi < n_F / 2. / lambda_F * sigma_w / sigma_F:
            print 'psi is too small, smalles value possible: ', n_F / 2. / lambda_F * sigma_w / sigma_F
            return psi, psi
        RP = 10000
        Rand = (np.random.rand(RP) - .5) * dF
        psi = [psi]
        len_psi = 1
    C = np.arange(n_F**2).reshape(n_F, n_F)
    # xi = 1.
    Rand = (np.random.rand(RP) - .5) * dF
    Rand2 = (np.random.rand(RP) - .5) * dF

    for i in range(len_psi):
        print i + 1
        P = np.zeros((n_F**2, n_F**2))
        a = (8. * psi[i]**2 * lambda_F**2 / n_F**2 * (sigma_F**2 / sigma_w**2) - 2.) * xi
        b = 2. * psi[i] * lambda_F * sigma_F / n_F
        W = b * np.array([-1, 0, 1])
        pW = 1. / (2 + a) * np.array([1, a, 1])
        c = 0
        for F1_idx in xrange(n_F):
            for F2_idx in xrange(n_F):
                for W_idx in xrange(3):
                    F1 = np.repeat(F[F1_idx], RP) + Rand
                    F2 = np.repeat(F[F2_idx], RP) + Rand2
                    F1_mapped = F1 * AR1 + F2 * AR2 + W[W_idx]
                    F2_mapped_idx = np.repeat(F1_idx, RP)
                    flim = F[-1] + dF / 2 * .99
                    F1_mapped = limit_F(flim, F1_mapped)
                    F1_mapped_idx = find_index(F1_mapped + dF / 2 * (n_F % 2), dF) + n_F / 2
                    cn = C[F1_mapped_idx, F2_mapped_idx]
                    for m in xrange(RP):
                        P[cn[m], c] += 1. * pW[W_idx] / float(RP)
                c += 1
        M = np.matrix(P)
        p = np.array((M**100000).T[0])[0]
        PFF = p.reshape(n_F, n_F)
        PF1 = np.sum(PFF[:, np.arange(n_F)], axis=1)
        PF2 = np.sum(PFF[np.arange(n_F), :], axis=0)
        f = np.linspace(- lambda_F * sigma_F, lambda_F * sigma_F, 1000)
        pF = ss.norm.pdf(f, loc=0, scale=sigma_F)
        if mode == 1:
            from mpl_toolkits.mplot3d import Axes3D
            fig1, ax1 = plt.subplots(1)
            fig2, ax2 = plt.subplots(subplot_kw=dict(projection='3d'))
            print 'alpha = ', a
            ax1.plot(F, PF1 / dF, '.-', c='b', label='$\psi$ = %s' % (psi[i]))
            ax1.plot(F, PF2 / dF, '.-', c='b', label=r'$\xi$ = %s' % (xi))
            ax1.plot(f, pF, label='exact', c='k')
            F1, F2 = np.meshgrid(F, F)
            surf = ax2.plot_surface(F1, F2, PFF, rstride=1, cstride=1, cmap='jet', alpha=1., linewidth=0.1, antialiased=True, shade=False)
            surf.set_clim(vmin=0, vmax=PFF.max())
            ax2.set_xlabel(r'F_{i}')
            ax2.set_ylabel(r'F_{i-1}')
            plt.legend(fancybox=True)
            plt.show()
            return F, PF1
        Fc = interp1d(F, PF1 / dF, kind='linear')
        error = np.sum((Fc(f) - pF)**2)**.5
        y[i] = error
    Rd = y
    R = interp1d(psi, Rd, kind='cubic')
    x = np.linspace(psi[0], psi[-1], 1000)
    psi_opt = np.argmin(R(x))
    psi_opt = round(x[psi_opt], 2)
    print 'optimal choice of $\psi$ =', psi_opt
    print 'alpha = ', a
    ax1.plot(x, R(x), '-')
    ax1.plot(psi, Rd, '.')
    ax1.plot(psi_opt, R(psi_opt), 'ro', label=r'$\psi = $%s' % (psi_opt))
    ax1.set_xlabel('$\psi$', size=15)
    ax1.set_ylabel('Residual', size=14)

    xi = np.linspace(xi / K, xi * K, iterations)
    y = np.zeros(len(xi))
    psi = psi_opt
    for i in range(len(xi)):
        print i + 1
        P = np.zeros((n_F**2, n_F**2))
        a = (8. * psi**2 * lambda_F**2 / n_F**2 * (sigma_F**2 / sigma_w**2) - 2.) * xi[i]
        b = 2. * psi * lambda_F * sigma_F / n_F
        W = b * np.array([-1, 0, 1])
        pW = 1. / (2 + a) * np.array([1, a, 1])
        c = 0
        for F1_idx in xrange(n_F):
            for F2_idx in xrange(n_F):
                for W_idx in xrange(3):
                    F1 = np.repeat(F[F1_idx], RP) + Rand
                    F2 = np.repeat(F[F2_idx], RP) + Rand2
                    F1_mapped = F1 * AR1 + F2 * AR2 + W[W_idx]
                    F2_mapped_idx = np.repeat(F1_idx, RP)
                    flim = F[-1] + dF / 2 * .99
                    F1_mapped = limit_F(flim, F1_mapped)
                    F1_mapped_idx = find_index(F1_mapped + dF / 2 * (n_F % 2), dF) + n_F / 2
                    cn = C[F1_mapped_idx, F2_mapped_idx]
                    for m in xrange(RP):
                        P[cn[m], c] += 1. * pW[W_idx] / float(RP)
                c += 1
        M = np.matrix(P)
        p = np.array((M**100000).T[0])[0]
        PFF = p.reshape(n_F, n_F)
        PF1 = np.sum(PFF[:, np.arange(n_F)], axis=1)
        PF2 = np.sum(PFF[np.arange(n_F), :], axis=0)
        f = np.linspace(- lambda_F * sigma_F, lambda_F * sigma_F, 1000)
        pF = ss.norm.pdf(f, loc=0, scale=sigma_F)


        Fc = interp1d(F, PF1 / dF, kind='linear')
        P = Fc(f)
        P = P/np.sum(P)
        std_F_hat = np.sum(f**2*P)**.5
        y[i] = std_F_hat
    Rd = (y - sigma_F)**2
    R = interp1d(xi, Rd, kind='cubic')
    x = np.linspace(xi[0], xi[-1], 1000)
    xi_opt = np.argmin(R(x))
    xi_opt = round(x[xi_opt], 2)
    print 'optimal choice of $xi$ =', xi_opt
    ax2.plot(x, R(x), '-')
    ax2.plot(xi, Rd, '.')
    ax2.plot(xi_opt, R(xi_opt), 'ro', label=r'$\xi = $%s' %(xi_opt))
    ax2.set_xlabel(r'$\xi$', size=15)
    ax2.set_ylabel('Residual', size=14)
    ax2.legend(fancybox=True)
    ax1.legend(fancybox=True)
    plt.tight_layout()
    plt.show()


def find_f_var():
    import matplotlib.pyplot as plt
    AR, sigma_w, F0, phi = cell.load_par()
    AR1 = AR[0]
    sigma_F = sigma_w / (np.sqrt(1. - AR1**2))
    dt, n_F, n_r, n_phi, psi, lambda_F, lambda_r = cell.discretization()
    F = np.linspace(- lambda_F * sigma_F, lambda_F * sigma_F, n_F)
    dF = F[1] - F[0]
    RP = 10000
    Rand = (np.random.rand(RP) - .5) * dF
    y = []
    x = []
    for i in range(10):
        g = i*0.05 + 0.8
        a = (8. * psi**2 * lambda_F**2 / (n_F**2 * (1. - AR[0]**2)) - 2.) * g
        b = 2. * psi * lambda_F * sigma_w / (n_F * np.sqrt(1. - AR[0]**2)) 
        W = b * np.array([-1, 0, 1])
        pW = 1. / (2 + a) * np.array([1, a, 1])
        print i
        P = np.zeros((n_F, n_F))
        for F_idx in xrange(n_F):
            for W_idx in xrange(3):
                F0 = np.repeat(F[F_idx], RP) + Rand
                F1 = F0 * AR1 + W[W_idx]
                flim = F[-1] + dF / 2 * .99
                F1 = limit_F(flim, F1)
                F1_idx = find_index(F1 + dF / 2 * (n_F % 2), dF) + n_F / 2
                for m in xrange(RP):
                    P[F1_idx[m], F_idx] += 1. * pW[W_idx] / float(RP)
        M = np.matrix(P)
        p = np.array((M**100000).T[0])[0]
        std_F = np.sum(F**2*p)**.5
        y.append(std_F/sigma_F)
        std_W = (W**2*pW)**.5
        x.append(g)
    fig, ax = plt.subplots()
    ax.plot(x, y)
    ax.hlines(1, x[0], x[-1])
    plt.show()





def image_map(X, V, F0, F1, F2, RKNpar, xlim, vlim, flim=0, zero_vel='no'):
	xn, vn = RKN(RKNpar, X, V, F0, F1, F2)  # Runge-Kutta-Nystrom
	Wall(X, V, xlim, vlim, xn, vn)  # Limit the velocity and displacement
	if flim != 0:
		fn = limit_F(flim, F2)
	else:
		fn = F2
	if zero_vel == 'yes':
		Wall_zero_vel(X, V, xn, vn)
		Wall(X, V, xlim, vlim, xn, vn)  # Limit the velocity and displacement
    
	return xn, vn, fn

def Pmatrix(mesh, par, cells='all', Rand_P=10, zero_vel='no', print_progress='no'):
    '''
    cells : which cells to map
    Rand_P : number of mc sim in each cell
    zero_vel : 'yes', 'no' : if buondry should be set to avoid zero velocity crossing
    mesh : [x, v, F] arrays tho specify centerpoints in cells. Default: get from cell_discretization
    par :  [omega_n, zeta, k, dt, AR1, sigma_w] defining parameters. Default: get from cell_discretization
    AR_noise term is always taken from cell_discretization -> discreization must be correct defined
    '''
    LC = cell.load_case(par)
    print 'Load case: ', LC
    dt = par['dt']
    x = mesh['x']
    v = mesh['v']
    dx = x[1] - x[0]
    dv = v[1] - v[0]
    size = [x.size, v.size]
    ar_load = False
    harm_load = False
    flim = 0
    if LC in [1, 3, 4, 6, 7, 8]:   # If ther is an ar load component in the load case
    	ar_load = True
    	F = mesh['F']
    	AR = par['AR']
    	sigma_w = par['sigma_w']
    	dF = F[1] - F[0]
    	n_F = len(F)
    	flim = F[-1] + (F[1] - F[0]) / 2. * .99
    	for i in xrange(len(AR)):
    		size.append(F.size)
    	if AR.all() == 0:

            size_w = n_F
            W = np.linspace(-3 * sigma_w, 3 * sigma_w, size_w)
            dW = W[1] - W[0]
            WU, WL = W + 0.5 * dW, W - 0.5 * dW
            pW = ss.norm.cdf(WU, loc=0, scale=sigma_w) - ss.norm.cdf(WL, loc=0, scale=sigma_w)
            pW = pW / np.sum(pW)
        else:
            W, pW = cell.AR_NOISE()
            size_w = 3
    if LC in [2, 4, 5, 7, 8, 9]:		# If there is a harmonic load component
    	harm_load = True
    	phi = mesh['phi']
    	F0 = par['F0']
    	omega_f = par['omega_f']
    	dphi = phi[1] - phi[0]
    	size.append(phi.size)
    	if LC in [5, 8, 9]:
    		phi2 = mesh['phi2']
    		size.append(phi2.size)
    		dphi2 = phi2[1] - phi2[0]
    ndim = len(size)
    RKNpar = par['omega_n'], par['zeta'], par['k'], dt
    ncells = 1
    for i in range(ndim):
        ncells = ncells * size[i]
    C = np.arange(ncells, dtype=np.uint32).reshape(size)
    n_it = 1e8  # number if iterations in each chunk (hardware dependend) (chunks might be solved in parallel!)
    if ar_load:
        Rand_P = Rand_P * size_w
    if type(cells) is str:  # When all cells are mapped
        numthreads = int(ncells * Rand_P // n_it + 1)
        chunklen = ncells / numthreads
    else:  # When a list of cells are mapped
        numthreads = int(cells.size * Rand_P // n_it + 1)
        chunklen = cells.size / numthreads
    c = np.arange(ncells, dtype=np.uint32)
    if type(cells) is str:
        if cells == 'all':
            cells = c
    chunks = [cells[i * chunklen:min([cells[-1] + 1, (i + 1) * chunklen])] for i in range(numthreads)]
    print "Number of chunks: ", len(chunks)
    print 'Number of iterations to be solved in millions: ', round(ncells * Rand_P / 1e6, 2)
    rx = np.random.rand(chunklen * Rand_P) - .5
    rv = np.random.rand(chunklen * Rand_P) - .5
    rk = [np.random.rand(chunklen * Rand_P) - .5 for i in xrange(ndim - 2)]

    xlim = size[0] * dx / 2. - .1e-3 * dx
    vlim = size[1] * dv / 2. - .1e-3 * dv
    t0 = time.time()
    it = 0    # number of iterations
    i = 0     # numbers of loop while t < t0
    tt0 = time.time()
    P = csc_matrix((ncells, ncells))
    i = 0
    for chunk in chunks:
        print 'Calculating chunck no. %s of %s' % (i + 1, len(chunks))
        t0 = time.time()
        n = len(chunk) * Rand_P
        if ndim == 3:
        	X, V, K1 = argwhere3d(size[1], size[2], chunk)
        elif ndim == 4:
        	X, V, K1, K2 = argwhere4d(size[1], size[2], size[3], chunk)
        elif ndim == 5:
        	X, V, K1, K2, K3 = argwhere5d(size[1], size[2], size[3], size[4], chunk)

        X = np.repeat(X, Rand_P)
        V = np.repeat(V, Rand_P)
        X = index_to_value(X, x, rx[:n])
        V = index_to_value(V, v, rv[:n])


        if LC == 1: 
            f0 = np.repeat(K1, Rand_P)
            f0 = index_to_value(f0, F, rk[0][:n])
            f2 = np.tile(W, n / size_w) + f0 * AR[0]
            f1 = 0.5 * (f0 + f2)
            xn, vn, K1n = image_map(X, V, f0, f1, f2, RKNpar, xlim, vlim, flim, zero_vel)
            xn_idx = find_index(xn + dx / 2 * (size[0] % 2), dx) + size[0] / 2
            vn_idx = find_index(vn + dv / 2 * (size[1] % 2), dv) + size[1] / 2
            K1n_idx = find_index(K1n + dF / 2 * (n_F % 2), dF) + n_F / 2
            cn = C[xn_idx, vn_idx, K1n_idx]
            data_i = np.tile(pW, n / size_w) / Rand_P * size_w

        elif LC ==2:
            phi0 = np.repeat(K1, Rand_P)
            phi0 = index_to_value(phi0, phi, rk[0][:n])
            f0 = F0[0] * np.sin(phi0 )
            f1 = F0[0] * np.sin(phi0 + omega_f[0] * dt / 2 )
            f2 = F0[0] * np.sin(phi0 + omega_f[0] * dt )
            xn, vn, f2 = image_map(X, V, f0, f1, f2, RKNpar, xlim, vlim, flim, zero_vel)
            xn_idx = find_index(xn + dx / 2 * (size[0] % 2), dx) + size[0] / 2
            vn_idx = find_index(vn + dv / 2 * (size[1] % 2), dv) + size[1] / 2
            K1n = (phi0 + omega_f[0] * dt * N) % (2 * np.pi)
            K1n_idx = np.int_(K1n / dphi)
            cn = C[xn_idx, vn_idx, K1n_idx]
            data_i = np.ones(n) / Rand_P

        elif LC == 3:
            f0 = np.repeat(K1, Rand_P)
            fneg1 = np.repeat(K2, Rand_P)
            f0 = index_to_value(f0, F, rk[0][:n])
            fneg1 = index_to_value(fneg1, F, rk[1][:n])
            f2 = np.tile(W, n / size_w) + f0 * AR[0] + fneg1 * AR[1]
            f1 = 0.5 * (f0 + f2)
            xn, vn, K1n = image_map(X, V, f0, f1, f2, RKNpar, xlim, vlim, flim, zero_vel)
            xn_idx = find_index(xn + dx / 2 * (size[0] % 2), dx) + size[0] / 2
            vn_idx = find_index(vn + dv / 2 * (size[1] % 2), dv) + size[1] / 2
            K2n = f0
            K1n_idx = find_index(K1n + dF / 2 * (n_F % 2), dF) + n_F / 2
            K2n_idx = find_index(K2n + dF / 2 * (n_F % 2), dF) + n_F / 2
            cn = C[xn_idx, vn_idx, K1n_idx, K2n_idx]
            data_i = np.tile(pW, n / size_w) / Rand_P * size_w

        elif LC == 4:
            f0 = np.repeat(K1, Rand_P)
            f0 = index_to_value(f0, F, rk[0][:n])
            f2 = np.tile(W, n / size_w) + f0 * AR[0]
            f1 = 0.5 * (f0 + f2)
            K1n = limit_F(flim, f2)
            K1n_idx = find_index(K1n + dF / 2 * (n_F % 2), dF) + n_F / 2
            phi0 = np.repeat(K2, Rand_P)
            phi0 = index_to_value(phi0, phi, rk[1][:n])
            f0 += F0[0] * np.sin(phi0)
            f1 += F0[0] * np.sin(phi0 + omega_f[0] * dt / 2)
            f2 += F0[0] * np.sin(phi0 + omega_f[0] * dt)
            xn, vn, fn = image_map(X, V, f0, f1, f2, RKNpar, xlim, vlim, flim, zero_vel)
            xn_idx = find_index(xn + dx / 2 * (size[0] % 2), dx) + size[0] / 2
            vn_idx = find_index(vn + dv / 2 * (size[1] % 2), dv) + size[1] / 2
            K2n = (phi0 + omega_f[0] * dt ) % (2 * np.pi)
            K2n_idx = np.int_(K2n / dphi)
            cn = C[xn_idx, vn_idx, K1n_idx, K2n_idx]
            data_i = np.tile(pW, n / size_w) / Rand_P * size_w

        elif LC == 5:
            phi0 = np.repeat(K1, Rand_P)
            phi0 = index_to_value(phi0, phi, rk[0][:n])
            K1n = (phi0 + omega_f[0] * dt ) % (2 * np.pi)
            K1n_idx = np.int_(K1n / dphi)

            phi1 = np.repeat(K2, Rand_P)
            phi1 = index_to_value(phi1, phi2, rk[1][:n])
            K2n = (phi1 + omega_f[1] * dt ) % (2 * np.pi)
            K2n_idx = np.int_(K2n / dphi2)
            f0 = F0[0] * np.sin(phi0) +                        F0[1] * np.sin(phi1)
            f1 = F0[0] * np.sin(phi0  + omega_f[0] * dt / 2) + F0[1] * np.sin(phi1 + omega_f[1] * dt / 2)
            f2 = F0[0] * np.sin(phi0  + omega_f[0] * dt)     + F0[1] * np.sin(phi1 + omega_f[1] * dt)
            xn, vn, fn = image_map(X, V, f0, f1, f2, RKNpar, xlim, vlim, flim, zero_vel)
            xn_idx = find_index(xn + dx / 2 * (size[0] % 2), dx) + size[0] / 2
            vn_idx = find_index(vn + dv / 2 * (size[1] % 2), dv) + size[1] / 2
            cn = C[xn_idx, vn_idx, K1n_idx, K2n_idx]
            data_i = np.ones(n) / Rand_P



        col_i = np.repeat(chunk, Rand_P)  # column idx
        row_i = cn.flatten()  # row idx
        P_i = csc_matrix((data_i, (row_i, col_i)), shape=(ncells, ncells))
        P += P_i
        it += n
        t = time.time() - t0
        i += 1
        print 'Time used calculating chunk: %s seconds.' %(round(t, 2))
    if print_progress != 'no':
        print 'creating random took :'
        print 'RKN and limiting took: '
        print 'INDEXING took: '
        print
    print 'loop took: ', time.time() - tt0
    print 'Monte carlo points in each cell: ', int(Rand_P)
    print 'Step took %s seconds' % (time.time() - t0)
    print
    print 'Total iterations done in millions: ', it * 1e-6
    return P


def boundary_matrix(P, Rand_P, mesh, par):
    '''
    P : probability matrix
    Rand_P : number of simulations for boundary cells
    mesh : x, v, F : arrays of centerpoints in cells
    par : [omega_n, zeta, k, dt, AR1, sigma_w]
    '''
    def find_crossing_cells(P, mesh):
        ndim = len(mesh.keys())
        x = mesh['x']
        v = mesh['v']
        del mesh['x']
        del mesh['v']
        size = [x.size, v.size]
        for key in mesh.keys():
        	size.append(mesh[key].size)
        mesh['x'] = x
        mesh['v'] = v
        ncells = np.prod(size)
        v_neg = np.arange(size[1] / 2 - 1 + size[1] % 2)
        v_pos = np.arange(size[1] / 2 + 1, size[1])

        C = np.arange(ncells).reshape(size)
        if ndim == 3:
        	cneg = C[:, v_neg, :].flatten()
        	cpos = C[:, v_pos, :].flatten()
        elif ndim == 4:
        	cneg = C[:, v_neg, :, :].flatten()
        	cpos = C[:, v_pos, :, :].flatten()

        c_sign = np.zeros(ncells, dtype=np.int8)
        c_sign[cneg] = -1
        c_sign[cpos] = 1
        row_ind = P.indices
        col_ptr = P.indptr

        boundry_cells = []   # Cells states that crosses zeroe vel. in mapping
        for i in cneg:  # looping through all negativ velocity states
            cn = row_ind[col_ptr[i]:col_ptr[i + 1]]
            if (c_sign[cn] == -1).all():
                pass
            else:
                boundry_cells.append(i)

        for i in cpos:  # looping through all positive velocity states
            cn = row_ind[col_ptr[i]:col_ptr[i + 1]]
            if (c_sign[cn] == 1).all():
                pass
            else:
                boundry_cells.append(i)
        print 'crossing cells in Matrix [percent]: ', round(len(boundry_cells) / float(ncells) * 100., 2)
        return np.array(boundry_cells)

    def find_zero_cells(mesh):
        '''
        returns cells where velocity is zero
        '''
        ndim = len(mesh.keys())
        x = mesh['x']
        v = mesh['v']
        del mesh['x']
        del mesh['v']
        size = [x.size, v.size]
        for key in mesh.keys():
        	size.append(mesh[key].size)
        mesh['x'] = x
        mesh['v'] = v
        ncells = np.prod(size)
        v_zero = np.arange(size[1] / 2 - 1 + size[1] % 2, size[1] / 2 + 1)
        C = np.arange(ncells).reshape(size)
        if ndim == 3:
        	c_zeros = C[:, v_zero, :].flatten()
        elif ndim == 4:
        	c_zeros = C[:, v_zero, :, :].flatten()
        return c_zeros

    def Identity(n, c):
        '''
        Creates an identity matrix with n elements and ones placed at c
        '''
        cols = c
        rows = c
        data = np.ones(c.size)
        return csc_matrix((data, (rows, cols)), shape=(n, n), dtype=np.int8)

    n, n = P.shape
    print 'finding crossing cells...'
    b = find_crossing_cells(P, mesh)
    Ib = Identity(n, np.delete(np.arange(n), b))
    z = find_zero_cells(mesh)
    Iz = Identity(n, np.delete(np.arange(n), z))
    Izb = Identity(n, z)
    Pb = Pmatrix(mesh, par, cells=b, Rand_P=Rand_P, zero_vel='yes')
    return (Pb + P * Ib) * Iz + Izb
