# # -*- coding: utf-8 -*-
"""
Created on Sat Apr 1 12:24:58 2017

@author: Odd Eiken
"""
import numpy as np
import matplotlib.pyplot as plt
#%%
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cbook
from matplotlib import cm
from matplotlib.colors import LightSource
from matplotlib.ticker import LinearLocator, FormatStrFormatter
#%%
import scipy.stats as ss
from scipy.ndimage import gaussian_filter
from main import load_array, load_sparse_csc
from cell_mapping import boundary_matrix


def data(p, mesh, smoothen=0):
    ndim = len(mesh.keys())
    x = mesh['x']
    v = mesh['v']
    del mesh['x']
    del mesh['v']
    size = [x.size, v.size]
    for key in mesh.keys():
        size.append(mesh[key].size)
    mesh['x'] = x
    mesh['v'] = v
    ncells = np.prod(size)
    Q = p.reshape(size)
    if ndim == 4:
        Q = np.sum(Q[:, :, :, np.arange(size[3])], axis=3)
    PXV = np.sum(Q[:, :, np.arange(size[2])], axis=2)
    PXV = gaussian_filter(PXV, smoothen)
    PX = np.sum(PXV[:, np.arange(size[1])], axis=1)
    PV = np.sum(PXV[np.arange(size[0]), :], axis=0)
    return PX, PV, PXV.T


def plot3D(p, mesh, smoothen=0):
    PX, PV, PXV = data(p, mesh)
    x = mesh['x']
    v = mesh['v']
    dx = x[1] - x[0]
    dv = v[1] - v[0]
    Z = gaussian_filter(PXV / (dx * dv), smoothen)

    X, V = np.meshgrid(x, v)
    fig, ax = plt.subplots(subplot_kw=dict(projection='3d'),figsize=(10,10))
    surf = ax.plot_surface(V, X ,  Z, rstride=1, cstride=1, cmap='jet', alpha=1., linewidth=0., antialiased=True, shade=True)
    surf.set_clim(vmin=0, vmax=Z.max())
    ax.set_title('Joint probability Density Plot')
    ax.set_xlabel('Velocity')
    ax.set_ylabel('Displacement')


    
def force(p, mesh):
    ndim = len(mesh.keys())
    x = mesh['x']
    v = mesh['v']
    del mesh['x']
    del mesh['v']
    size = [x.size, v.size]
    for key in mesh.keys():
        size.append(mesh[key].size)
    mesh['x'] = x
    mesh['v'] = v
    ncells = np.prod(size)
    Q = p.reshape(size)
    if ndim == 4:
        Q = np.sum(Q[:, :, :, np.arange(size[3])], axis=3)
    PXF = np.sum(Q[:, np.arange(size[1]), :], axis=1)
    PF = np.sum(PXF[np.arange(size[0]), :], axis=0)
    F = mesh['F']
    dF = F[1] - F[0]
    fig, ax = plt.subplots(figsize=(10, 7))
    ax.plot(F, PF / dF)
    ax.legend(fancybox=True)
    ax.set_xlabel("Newton")


def colormap(p, mesh, interpolation='nearest', smoothen=0, ndig=2, vmax=0):
    x = mesh['x']
    v = mesh['v']
    dx = x[1] - x[0]
    dv = v[1] - v[0]
    PX, PV, PXV = data(p, mesh)
    fig, ax = plt.subplots(figsize=(10, 7))
    Z = gaussian_filter(PXV / (dx * dv), smoothen)

    lim = [x[0], x[-1], v[0], v[-1]]
    ax.set_xlabel('Displacement')
    ax.set_ylabel('Velocity')
    if vmax == 0:
    	vmax = Z.max()
    cax = ax.imshow(Z, aspect='auto', interpolation=interpolation, cmap='jet', extent=lim, vmax=vmax, vmin=0)
    cbar = fig.colorbar(cax, ticks=[0, vmax / 2, vmax])
    c_max = round(vmax, ndig)
    c_mid = c_max / 2.
    cbar.ax.set_yticklabels(['0.0', c_mid, c_max])  # vertically oriented colorbar


def distribution(p, mesh, sigma_x=0, sigma_v=0, smoothen=0):
    x = mesh['x']
    v = mesh['v']
    dx = x[1] - x[0]
    dv = v[1] - v[0]
    PX, PV, PXV = data(p, mesh, smoothen=smoothen)
    fig, (ax1, ax2) = plt.subplots(2, figsize=(8, 6))
    ax1.set_title('Displacement')
    ax1.plot(x, PX / dx, '-', c='b', label='Cell Mapping')
    ax2.set_title('Velocity')
    ax2.plot(v, PV / dv, '-', c='r', label='Cell Mapping')
    if sigma_x != 0 and sigma_v != 0:
        x_ax = np.linspace(x[0], x[-1], 1000)
        pdf = ss.norm.pdf(x_ax, loc=0, scale=sigma_x)
        ax1.plot(x_ax, pdf, '--', c='k', lw=2, alpha=.3, label='Exact distr.')
        v_ax = np.linspace(v[0], v[-1], 1000)
        pdf = ss.norm.pdf(v_ax, loc=0, scale=sigma_v)
        ax2.plot(v_ax, pdf, '--', c='k', lw=2, alpha=.3, label='Exact distr.')
    ax2.legend(fancybox=True)
    ax1.legend(fancybox=True)


def multiple_plot(files, sigma_x=0, sigma_v=0):
    '''
    list of files
    '''
    fig, (ax1, ax2) = plt.subplots(2, figsize=(8, 6))
    ax1.set_title('Displacement')
    ax2.set_title('Velocity')
    for file in files:
        x, v, F, par, p = load_array(file)
        mesh = x, v, F
        dx = x[1] - x[0]
        dv = v[1] - v[0]
        PX, PV, PXV = data(p, mesh)
        ax1.plot(x, PX, '-', label=file)
        ax2.plot(v, PV, '-', label=file)
        varx = np.sum(x**2 * PX) * dx
        print varx**.5
    if sigma_x != 0 and sigma_v != 0:
        x_ax = np.linspace(x[0], x[-1], 1000)
        pdf = ss.norm.pdf(x_ax, loc=0, scale=sigma_x)
        ax1.plot(x_ax, pdf, '--', c='k', lw=2, alpha=.3, label='Exact distr.')
        v_ax = np.linspace(v[0], v[-1], 1000)
        pdf = ss.norm.pdf(v_ax, loc=0, scale=sigma_v)
        ax2.plot(v_ax, pdf, '--', c='k', lw=2, alpha=.3, label='Exact distr.')
    ax1.legend(fancybox=True)
    ax2.legend(fancybox=True)
    # ax1.axis([-3, 3, 0, .5])
    # ax2.axis([-6, 6, 0, .25])
    plt.show()


def plot_halfcycle(filename, x0):
    x, v, F, par, P_init = load_sparse_csc(filename)
    mesh = x, v, F
    # Pb = boundary_matrix(P_init, 10, mesh=mesh, par=par)
    ncells = x.size * v.size * F.size
    v0_idx = np.argmin(abs(v))
    cx = np.arange(ncells).reshape(x.size, v.size, F.size)[:, v0_idx, :].flatten()  # cells along x - axis
    x0_idx = np.argmin(np.abs(x - x0))
    c0 = np.arange(ncells).reshape(x.size, v.size, F.size)[x0_idx, v0_idx, :].flatten()  # cells at starting point
    # p = starting_vector(Pb, 50, x0_idx, mesh=mesh, par=par)
    lim = [-3, 3, -3*2.1, 3*2.1]
    max_it = 4
    PXT = np.zeros((max_it, x.size))
    Pb = P_init
    p = np.zeros(ncells)
    p[c0] = 1. / len(c0)
    for i in xrange(max_it):
        p = Pb * p
    PX, PV, PXV = data(p, mesh)
    fig, ax = plt.subplots()
    ax.set_xlabel('Displacement')
    ax.set_ylabel('Velocity')
    ax.imshow(PXV, aspect='auto', interpolation='nearest', cmap='jet', extent=lim)
    plt.show()
    # colormap(p, mesh)


if __name__ == "__main__":
    from cell_discretization import FRF
    def prob_plot(X,x):
        x = np.asarray([x]).flatten()
        P = np.zeros(len(x))
        for i in xrange(len(x)):
            if x[i]**2>X**2:
                P[i]=0
            elif x[i]**2 == X**2:
                P[i] = np.inf
            else:
                P[i] = (np.pi*(X**2-x[i]**2)**.5)**-1
        return P

    G = FRF(3) * 1e5 * 3
    x = np.linspace(-G*1.2, G*1.2, 1000)
    p = prob_plot(G, x)
    sigma = 0.267202794578
    x =np.linspace(-4*sigma, 4*sigma, 1000)
    p = ss.norm.pdf(x, scale=sigma)
    fig, ax = plt.subplots(figsize=(10, 7))
    plt.plot(x, p, c='k', lw = 2, alpha=0.6, label='exact')   # harmonic distribution
    c = ['#33cc33', 'r', 'b', '#cc0099', '#cc6600', '#33ccff', '#9933ff']
    i = 0
    files = np.array(["$\Delta t=0.01s$", "$\Delta t=0.05s$", '$\Delta t=0.1s$', '$\Delta t=0.2s$', '$\Delta t=0.5s$'])
    labels = np.array(["1 RP", "10RP", '100 RP', '1000 RP'])
    labels = files
    which = 'v'
    smooth=np.ones(5) * 1
    for file in files:
        mesh, par, p = load_array(file)
        PX, PV, PXV = data(p, mesh, smoothen=smooth[i])
        x = mesh['x']
        dx = x[1] - x[0]
        v = mesh['v']
        dv = v[1] - v[0]
        if which == 'x':
            ax.plot(x, PX / dx, '-', label=labels[i], c=c[i], lw=1.2)
            std = (np.sum(x**2 * PX / dx) * dx)**.5
        else:
            ax.plot(v, PV / dv, '-', label=labels[i], c=c[i], lw=1.2)
            std = (np.sum(v**2 * PV / dv) * dv)**.5
        print 'STD: ', std
        print 'error:', std/.267202794578*100 - 100
        print 
        i += 1
    plt.legend(fancybox=True)
    plt.show()


